import ijeoma.motion.*;
import ijeoma.motion.tween.*;
import ijeoma.motion.easing.*;

import java.awt.Dimension;
import java.io.PrintStream;
import java.nio.*;
import java.util.ArrayList;

import processing.core.*;
import processing.opengl.*;

import controlP5.ControlEvent;
import controlP5.ControlP5;

import peasy.*;
import peasy.org.apache.commons.math.geometry.Vector3D;

import toxi.geom.*;
import toxi.math.MathUtils;
import toxi.processing.*;

import javax.media.opengl.*;
import javax.media.opengl.glu.*;

import ddf.minim.analysis.*;
import ddf.minim.*;

/**
 * 
 * @author mannytan
 * @name NudiBranch
 * @date Dec 11, 2010
 * 
 */
public class NudiBranch extends PApplet {

	// Fan fan;
	Guides guides;
	PeasyCam cam;

	int totalSegments = 33;
	Segment firstSegment;
	ArrayList<Segment> segmentList = new ArrayList();

	private static PrintStream out = System.out;

	public static ToxiclibsSupport gfx;

	private Vec3D mouseVec;

	Minim minim;
	AudioPlayer jingle;
	FFT fftLin;
	FFT fftLog;
	float height3;
	float height23;
	float spectrumScale = 2;
	float speeder = 0;

	// controls
	Controls controls;

	//*
	static public void main(String args[]) {
		// PApplet.main(new String[] {"--present", "NudiBranch" });
		PApplet.main(new String[] { "NudiBranch" });
	}

	//*/

	public void setup() {
		size(1200, 800, OPENGL);
		// hint(DISABLE_DEPTH_TEST);
		// hint(ENABLE_DEPTH_SORT);
		this.setSize(new Dimension(1200, 800)); // resize & set bounds of the PApplet

		background(10);

		// Create Camera
		cam = new PeasyCam(this, 1000);
		cam.lookAt(0, 0, 0, 0);

		// Create ToxiclibsSupport
		gfx = new ToxiclibsSupport(this);

		controls = new Controls(this);
		controls.createControls();
		controls.setAsDefaultSliders();

		// Create Guides
		guides = new Guides(this);

		// Create first segment data
		firstSegment = new Segment(this);
		firstSegment.seed = 0;
		firstSegment.seedNormal = 0.0f;
		firstSegment.seedNormalReversed = 1.0f - firstSegment.seedNormal;
		firstSegment.parentSegment = firstSegment;

		// Create Segments
		for (int i = 0; i < totalSegments; i++) {
			Segment segment = new Segment(this);
			segment.seed = (i + 1);
			segment.seedNormal = (float) (i + 1) / (float) totalSegments;
			segment.seedNormalReversed = 1.0f - segment.seedNormal;

			// if first segment, make parent itself
			if (i == 0) {
				segment.parentSegment = firstSegment;
			} else {
				segment.parentSegment = segmentList.get(segmentList.size() - 1);
			}

			segment.randomizeCurrentControls();
			segment.setPreviousControls();
			segmentList.add(segment);
		}

		gl = ((PGraphicsOpenGL) g).gl;
		glu = ((PGraphicsOpenGL) g).glu;

		mouseVec = new Ray3D();

		// capture audio data
		minim = new Minim(this);

		// jingle = minim.loadFile("recognizer.mp3", 2048);
		jingle = minim.loadFile("derezzed.mp3", 2048);

		jingle.loop();
		fftLog = new FFT(jingle.bufferSize(), jingle.sampleRate());
		// fftLog.logAverages(22, 3); 
		// fftLog.avgSize() = 30;

		// fftLog.logAverages(60, 5);
		// fftLog.avgSize() = 45;

		fftLog.logAverages(60, 7);
		// fftLog.avgSize() = 63;
		// println(fftLog.avgSize());

	}

	public void draw() {

		controls.updateParams();
		// this.rotateY(frameCount*.01f);
		background(5);

		// camera HUD
		cam.beginHUD();
		guides.draw2DScope();
		controls.draw();
		updateControlParams();
		cam.endHUD();
		lights();

		// draw guides
		if (controls.isGuideVisible) {
			guides.draw3DAxes(400.0f);
		}
		guides.draw3DBox(400.0f);

		//*
		// create mouse pointer

		speeder += .025f * controls.rootSpeed;
		mouseVec.x = sin(MathUtils.PI * speeder * .5f) * controls.rootRadius + 0.0f;
		mouseVec.y = cos(MathUtils.PI * speeder) * controls.rootRadius + 0.0f;
		mouseVec.z = sin(MathUtils.PI * speeder) * controls.rootRadius + 0.0f;
		// */

		/*
		// mouse position
		float[] mousePosList = getMouse3D();
		mouseVec.x += (mousePosList[0] - mouseVec.x) * .05f;
		mouseVec.y += (mousePosList[1] - mouseVec.y) * .05f;
		mouseVec.z += (mousePosList[2] - mouseVec.z) * .05f;
		// mouseVec.normalizeTo(500.0f);
		// */

		/*
		// limiters
		if (mouseVec.z < -2000) {
			mouseVec.z = -2000;
		}
		if (mouseVec.z > 2000) {
			mouseVec.z = 2000;
		} 
		// */

		// follow point
		// setLineStyle(5, 0xFFFF6600, 1.0f);
		// gfx.point(mouseVec);

		// first segment
		firstSegment.current.set(mouseVec);
		firstSegment.previous.set(mouseVec);
		firstSegment.parse();

		Segment segment;
		segment = segmentList.get(0);
		segment.parse();

		int startId = 2;
		for (int i = 0; i < totalSegments; i++) {
			segment = segmentList.get(i);
			segment.parse();
		}

		if (controls.isFillMeshVisible) {
			for (int i = startId; i < totalSegments; i++) {
				segment = segmentList.get(i);
				segment.drawFillMesh(gfx);
			}
		}

		if (controls.isDotMeshVisible) {
			for (int i = startId; i < totalSegments; i++) {
				segment = segmentList.get(i);
				segment.drawDotMesh(gfx);
			}
		}
		if (controls.isLineTrailVisible) {
			for (int i = startId; i < totalSegments; i++) {
				segment = segmentList.get(i);
				segment.drawLineTrailMesh(gfx);
			}
		}
		if (controls.isLineHairVisible) {
			for (int i = startId; i < totalSegments; i++) {
				segment = segmentList.get(i);
				segment.drawLineHairMesh(gfx);
			}
		}
		if (controls.isLineCenterVisible) {
			for (int i = startId; i < totalSegments; i++) {
				segment = segmentList.get(i);
				segment.drawLineCenter(gfx);
			}
		}

	}

	public void keyPressed() {
		if (key == ' ') {
			reset();
		}
	}

	public void reset() {
		controls.randomizeSliders();

		/*
		for (Segment segment : segmentList) {
			segment.randomizeCurrentControls();
			segment.setPreviousControls();
		}
		*/
	}

	GL gl;
	GLU glu;

	float[] getMouse3D() {
		((PGraphicsOpenGL) g).beginGL();
		int viewport[] = new int[4];
		double[] proj = new double[16];
		double[] model = new double[16];
		gl.glGetIntegerv(GL.GL_VIEWPORT, viewport, 0);
		gl.glGetDoublev(GL.GL_PROJECTION_MATRIX, proj, 0);
		gl.glGetDoublev(GL.GL_MODELVIEW_MATRIX, model, 0);
		FloatBuffer fb = ByteBuffer.allocateDirect(4).order(ByteOrder.nativeOrder()).asFloatBuffer();
		gl.glReadPixels(mouseX, height - mouseY, 1, 1, GL.GL_DEPTH_COMPONENT, GL.GL_FLOAT, fb);
		fb.rewind();
		double[] mousePosArr = new double[4];
		glu.gluUnProject((double) mouseX, height - (double) mouseY, (double) fb.get(0), model, 0, proj, 0, viewport, 0, mousePosArr, 0);
		((PGraphicsOpenGL) g).endGL();
		return new float[] { (float) mousePosArr[0], (float) mousePosArr[1], (float) mousePosArr[2] };
	}

	private void updateControlParams() {

		fftLog.forward(jingle.mix);
		for (Segment segment : segmentList) {
			segment.segmentScalar = controls.rootScalar;
			segment.frameCountScalar = controls.rootFrameCountScalar;
			segment.seedNormalScalar = controls.rootSeedNormalScalar;

			segment.layerRadiusOffset = controls.rootLayerRadiusOffset;
			int avg = Math.abs(frameCount - segment.seed) % (totalSegments - 1);
			float layerRadiusScalar = (float) Math.sqrt((double) fftLog.getAvg(avg)) * controls.rootLayerAudioScalar + controls.rootLayerRadiusScalar;
			segment.layerRadiusScalar = layerRadiusScalar;

			segment.layerSpeedOffset = controls.rootLayerSpeedOffset;
			segment.layerRadiusOscillator = controls.rootLayerRadiusOscillator;
			segment.layerRadiusOscillatorOffset = controls.rootLayerRadiusOscillatorOffset;
			segment.flappingSpeedNormal = controls.rootFlappingSpeedNormal;
			segment.flappingSpeedOffset = controls.rootFlappingSpeedOffset;
			segment.maxAngle = controls.rootMaxAngle;
			segment.frameCount = frameCount;

		}
	}

	public void stopCamera() {
		System.out.println("stopCamera");
		// cam.setActive(false);
	}

	public void controlEvent(ControlEvent theEvent) {
		// cam.setActive(false);
	}

	public void mouseReleased() {
		cam.setActive(true);
	}

	public void setLineStyle(int weightParam, int colorParam, float alphaParam) {
		this.noFill();
		this.strokeWeight(weightParam);
		this.stroke(colorParam, 255.0f * alphaParam);
	}

	public float getRandom(float a, float b) {
		return (float) Math.random() * (b - a) + a;
	}

}
