import java.awt.Color;
import java.util.ArrayList;

import processing.core.PApplet;
import toxi.geom.*;
import toxi.geom.mesh.TriangleMesh;
import toxi.geom.mesh.TriangleMesh.*;
import toxi.math.MathUtils;
import toxi.processing.ToxiclibsSupport;

public class Segment extends BoilerPlate {
	PApplet parent;
	float degreeToRadMultiplier = MathUtils.PI / 180;

	public int seed = 0;
	public float seedNormal = 0.0f;
	public float seedNormalReversed = 1.0f;

	public Ray3D currentRotator;
	public Ray3D currentArbitrary;
	public Ray3D currentTangent;

	public Vec3D currentDirection;
	public Ray3D current;

	public Vec3D previousDirection;
	public Ray3D previous;

	public Segment parentSegment;

	// static params
	public int totalWings = 2;
	public int totalLayers = 30;

	// dynamic params
	public float frameCountScalar;
	private float frameCountScalarDelta = 0;
	public float segmentScalar;
	public float seedNormalScalar;
	public float layerRadiusOffset;
	public float layerRadiusScalar;
	public float layerSpeedOffset;
	public float layerRadiusOscillator;
	public float layerRadiusOscillatorOffset;

	public float flappingSpeedOffset;
	public float flappingSpeedNormal;
	private float flappingSpeedNormalDelta = 0;
	public float maxAngle;
	
	TriangleMesh segmentMesh;
	TriangleMesh[][] squareMeshList = new TriangleMesh[totalWings][totalLayers];

	public Vec3D[][] segmentPoints = new Vec3D[totalWings][totalLayers];
	public int frameCount;

	Segment(PApplet p) {
		parent = p;
		setScope("Segment");
		currentDirection = new Vec3D();
		current = new Ray3D();
		previousDirection = new Vec3D();
		previous = new Ray3D();

		currentArbitrary = new Ray3D(new Vec3D(), Vec3D.Y_AXIS);
		currentTangent = new Ray3D();
		currentRotator = new Ray3D();

		int i, j;
		for (i = 0; i < totalWings; i++) {
			for (j = 0; j < totalLayers; j++) {
				segmentPoints[i][j] = new Vec3D();
			}
		}

		segmentMesh = new TriangleMesh();
		Vec3D a, b, c, d;
		float rad = 1.0f;
		for (i = 0; i < totalWings; i++) {
			for (j = 1; j < totalLayers; j++) {
				a = new Vec3D(getRandom(-rad, rad), getRandom(-rad, rad), getRandom(-rad, rad));
				b = new Vec3D(getRandom(-rad, rad), getRandom(-rad, rad), getRandom(-rad, rad));
				c = new Vec3D(getRandom(-rad, rad), getRandom(-rad, rad), getRandom(-rad, rad));
				d = new Vec3D(getRandom(-rad, rad), getRandom(-rad, rad), getRandom(-rad, rad));

				segmentMesh.addFace(a, b, c);
				segmentMesh.addFace(d, c, b);
			}
		}

		for (i = 0; i < totalWings; i++) {
			for (j = 1; j < totalLayers; j++) {
				a = new Vec3D(getRandom(-rad, rad), getRandom(-rad, rad), getRandom(-rad, rad));
				b = new Vec3D(getRandom(-rad, rad), getRandom(-rad, rad), getRandom(-rad, rad));
				c = new Vec3D(getRandom(-rad, rad), getRandom(-rad, rad), getRandom(-rad, rad));
				d = new Vec3D(getRandom(-rad, rad), getRandom(-rad, rad), getRandom(-rad, rad));

				squareMeshList[i][j] = new TriangleMesh();
				squareMeshList[i][j].addFace(a, b, c);
				squareMeshList[i][j].addFace(d, c, b);
			}
		}

	}

	public void randomizeCurrentControls() {
		float r;
		r = 1.0f;
		currentDirection = new Vec3D();
		r = 100.0f;
		current = new Ray3D();
	}

	public void setPreviousControls() {
		previousDirection = parentSegment.currentDirection;
		previous = parentSegment.current;
	}

	public void parse() {
		previousDirection.set(parentSegment.currentDirection);
		previous.set(parentSegment.current);

		current.set(current.sub(previous));
		current.normalizeTo(segmentScalar);
		current.addSelf(previous);
		current.setDirection(previous.sub(current));

		currentArbitrary.set(current);
		currentTangent.set(current);
		currentRotator.set(current);

		Vec3D currentVec = new Vec3D(current);
		Vec3D dirVec = new Vec3D(current.getDirection());
		Vec3D crossVec = currentArbitrary.cross(dirVec.getNormalized()).normalize();

		frameCountScalarDelta += frameCountScalar;
		flappingSpeedNormalDelta += flappingSpeedNormal;
		for (int i = 0; i < totalWings; i++) {
			for (int j = 0; j < totalLayers; j++) {
				// convert positions
				float wingPercentage = (float) i / (float) totalWings;
				float layerPercentage = (float) j / (float) totalLayers;

				float angleInRads = MathUtils.TWO_PI * wingPercentage;
				angleInRads += seedNormal * seedNormalScalar; // extra offset of rotation
				
				angleInRads += frameCountScalarDelta;
				
				float flappingSpeed = 0;
				flappingSpeed += flappingSpeedNormalDelta;
				flappingSpeed += seedNormal * flappingSpeedOffset;
				flappingSpeed += (1 - layerPercentage) * layerSpeedOffset;
				// flips it around based on 0,1 --> -1,1
				float flappingRange = maxAngle * (float) ((i % 2) * 2 - 1);

				// degreeToRadMultiplier
				float rangeOffsetInRads = MathUtils.sin(flappingSpeed) * flappingRange / 360.0f * MathUtils.PI;

				angleInRads += (rangeOffsetInRads * (layerPercentage));

				// create local rotation vector
				Vec3D rotVec = crossVec.getRotatedAroundAxis(dirVec, angleInRads);

				// extra offset of rotation
				float layerRadius = layerRadiusScalar * layerPercentage;
				layerRadius += layerRadiusOffset;
				layerRadius *= layerRadiusOscillator;

				Vec3D point = new Vec3D(rotVec);
				point.scaleSelf(layerRadius);
				point.addSelf(currentVec);

				segmentPoints[i][j].set(point);
			}
		}
	}

	public void drawLineCenter(ToxiclibsSupport gfx) {
		// tracer("draw");

		parent.noFill();
		float wingPercentage;
		float layerPercentage;

		// orange line down the center
		float colorAlpha = seedNormalReversed * 255.0f + 0.0f;
		setLineStyle(1, 0xFFFF6600, seedNormalReversed);
		gfx.ray(current, segmentScalar);

	}

	public void drawLineHairMesh(ToxiclibsSupport gfx) {
		parent.noFill();
		float wingPercentage;
		float layerPercentage;
		// hairs
		for (int i = 0; i < totalWings; i++) {
			for (int j = 1; j < totalLayers; j++) {
				wingPercentage = (float) i / (float) totalWings;
				layerPercentage = (float) j / (float) totalLayers;

				float lineAlpha = (seedNormalReversed) * (1 - layerPercentage);

				Vec3D a = new Vec3D(segmentPoints[i][j]);
				a.add(current);
				Vec3D b = new Vec3D(segmentPoints[i][j - 1]);
				b.add(previous);

				setLineStyle(1, 0xFF00FFFF, lineAlpha);
				gfx.line(a, b);
			}
		}
	}

	public void drawLineTrailMesh(ToxiclibsSupport gfx) {
		parent.noFill();
		float wingPercentage;
		float layerPercentage;

		// strands
		for (int i = 0; i < totalWings; i++) {
			for (int j = 0; j < totalLayers; j++) {
				wingPercentage = (float) i / (float) totalWings;
				layerPercentage = (float) j / (float) totalLayers;

				float lineAlpha = (seedNormalReversed) * (1 - layerPercentage);

				Vec3D a = new Vec3D(segmentPoints[i][j]);
				a.add(current);
				Vec3D b = new Vec3D(parentSegment.segmentPoints[i][j]);
				b.add(previous);

				setLineStyle(1, 0xFF00FFFF, lineAlpha);
				gfx.line(a, b);
			}
		}
	}

	public void drawDotMesh(ToxiclibsSupport gfx) {
		parent.noFill();
		float wingPercentage;
		float layerPercentage;

		// yellow dots
		for (int i = 0; i < totalWings; i++) {
			for (int j = 0; j < totalLayers; j++) {
				wingPercentage = (float) i / (float) totalWings;
				layerPercentage = (float) j / (float) totalLayers;

				Vec3D point = new Vec3D(segmentPoints[i][j]);
				float lineAlpha = (seedNormalReversed) * (1 - layerPercentage);

				// setLineStyle(1, 0xFFFF6600, lineAlpha);
				setLineStyle((int)(seedNormalReversed*2.0f)+1, 0xFFFFFFFF, seedNormalReversed);
				gfx.point(point);
			}
		}
	}

	public void drawFillMesh(ToxiclibsSupport gfx) {
		// tracer("draw");
		parent.noFill();
		float wingPercentage;
		float layerPercentage;
		//*
		Face face;
		Vec3D a = new Vec3D();
		Vec3D b = new Vec3D();
		Vec3D c = new Vec3D();
		Vec3D d = new Vec3D();

		int faceCount = 0;
		// int innerColor = 0xFF00CCFF;//0xFFFF00FF;
		// int outerColor = 0xFFFF00FF;//0xFF00CCFF;
		float frameCountFraction = (float) (frameCount+0xFFFF0000) * .1f;
		float hueA = (float)((frameCountFraction)+seedNormalReversed*100)%1000*.001f;
		float hueB = (float)((frameCountFraction + 300)+seedNormalReversed*100)%1000*.001f;

	    int innerColor = Color.HSBtoRGB(hueA, .75f, 1.0f);
	    int outerColor = Color.HSBtoRGB(hueB, .75f, 1.0f);

		// int innerColor = rgb;//0xFFFF00FF;
		// int outerColor = 0xFFFF00FF;//0xFF00CCFF;
		for (int i = 0; i < totalWings; i++) {
			for (int j = 1; j < totalLayers; j++) {
				wingPercentage = (float) i / (float) totalWings;
				layerPercentage = (float) j / (float) totalLayers;
				a = segmentPoints[i][j];
				b = segmentPoints[i][j - 1];
				c = parentSegment.segmentPoints[i][j];
				d = parentSegment.segmentPoints[i][j - 1];

				face = squareMeshList[i][j].faces.get(0);
				face.a.set(a);
				face.b.set(b);
				face.c.set(c);

				face = squareMeshList[i][j].faces.get(1);
				face.a.set(d);
				face.b.set(c);
				face.c.set(b);

				squareMeshList[i][j].computeFaceNormals();
				squareMeshList[i][j].computeVertexNormals();

				int interColor = parent.lerpColor(innerColor, outerColor, layerPercentage);
				setFillStyle(interColor, seedNormalReversed*2);
				// setFillStyle(1, 0xFF00FFFF, 1.0f-layerPercentage);
				gfx.mesh(squareMeshList[i][j], true);
			}
		}
		// */

	}

	public void setFillStyle(int colorParam, float alphaParam) {
		parent.noStroke();
		parent.strokeWeight(1);
		parent.fill(colorParam, 255.0f * alphaParam);
	}

	public void setLineStyle(int weightParam, int colorParam, float alphaParam) {
		parent.noFill();
		parent.strokeWeight(weightParam);
		parent.stroke(colorParam, 255.0f * alphaParam);
	}
}
